/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, SAMSUNG ELECTRONICS CO., LTD. and BRITISH BROADCASTING CORPORATION
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of SAMSUNG ELECTRONICS CO., LTD. nor the name of the BRITISH BROADCASTING CORPORATION
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TComPic.cpp
    \brief    picture class
*/

#include "TComPic.h"

// ====================================================================================================================
// Constructor / destructor / create / destroy
// ====================================================================================================================

TComPic::TComPic()
{
  m_apcPicSym         = NULL;
  m_apcPicYuv[0]      = NULL;
  m_apcPicYuv[1]      = NULL;
#if HHI_INTERP_FILTER
  m_apcPicYuv[2]      = NULL;
#endif
  m_pcPicYuvPred      = NULL;
  m_pcPicYuvResi      = NULL;

#ifdef VIDYO_SVC
  m_apcScaledRlPicYuv[0] = NULL;
  m_apcScaledRlPicYuv[1] = NULL;
  m_apcScaledRlPicYuv[2] = NULL;
#endif

  m_bReconstructed    = false;
}

TComPic::~TComPic()
{
}

Void TComPic::create( Int iWidth, Int iHeight, UInt uiMaxWidth, UInt uiMaxHeight, UInt uiMaxDepth, Bool bIsVirtual )
{
  m_apcPicSym     = new TComPicSym;  m_apcPicSym   ->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  if (!bIsVirtual)
  {
    m_apcPicYuv[0]  = new TComPicYuv;  m_apcPicYuv[0]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  }
  m_apcPicYuv[1]  = new TComPicYuv;  m_apcPicYuv[1]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );

#if HHI_INTERP_FILTER
  // buffer for filtered reconstructed pic
  m_apcPicYuv[2]  = new TComPicYuv;  m_apcPicYuv[2]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
#endif

#ifdef VIDYO_SVC
  m_apcScaledRlPicYuv[0] = new TComPicYuv;  m_apcScaledRlPicYuv[0]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  m_apcScaledRlPicYuv[1] = new TComPicYuv;  m_apcScaledRlPicYuv[1]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  m_apcScaledRlPicYuv[2] = new TComPicYuv;  m_apcScaledRlPicYuv[2]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
#endif

  return;
}

Void TComPic::destroy()
{
  if (m_apcPicSym)
  {
    m_apcPicSym->destroy();
    delete m_apcPicSym;
    m_apcPicSym = NULL;
  }

  if (m_apcPicYuv[0])
  {
    m_apcPicYuv[0]->destroy();
    delete m_apcPicYuv[0];
    m_apcPicYuv[0]  = NULL;
  }

  if (m_apcPicYuv[1])
  {
    m_apcPicYuv[1]->destroy();
    delete m_apcPicYuv[1];
    m_apcPicYuv[1]  = NULL;
  }

#if HHI_INTERP_FILTER
  if (m_apcPicYuv[2])
  {
    m_apcPicYuv[2]->destroy();
    delete m_apcPicYuv[2];
    m_apcPicYuv[2]  = NULL;
  }
#endif

#ifdef VIDYO_SVC
  if (m_apcScaledRlPicYuv[0])
  {
    m_apcScaledRlPicYuv[0]->destroy();
    delete m_apcScaledRlPicYuv[0];
    m_apcScaledRlPicYuv[0]  = NULL;
  }
  if (m_apcScaledRlPicYuv[1])
  {
    m_apcScaledRlPicYuv[1]->destroy();
    delete m_apcScaledRlPicYuv[1];
    m_apcScaledRlPicYuv[1]  = NULL;
  }
  if (m_apcScaledRlPicYuv[2])
  {
    m_apcScaledRlPicYuv[2]->destroy();
    delete m_apcScaledRlPicYuv[2];
    m_apcScaledRlPicYuv[2]  = NULL;
  }
#endif

}

#ifdef VIDYO_SVC

Void TComPic::xGetScaledRlPicTo( TComPic* pcPic )
{
  Int   x, y, tmp;

  Pel*  pRecRl    = getPicYuvRec()->getLumaAddr(); // this recon should include IBDI
  Int   iStrideRl = getStride();
  Int   iWidthRl  = getPicYuvRec()->getWidth();
  Int   iHeightRl = getPicYuvRec()->getHeight();

  Pel*  pRec      = pcPic->getScaledRlPicRec()->getLumaAddr();
  Int   iStride   = pcPic->getStride();
  Int   iWidth    = pcPic->getPicYuvRec()->getWidth();
  Int   iHeight   = pcPic->getPicYuvRec()->getHeight();
 
  Int*  pTmpBuff;
  Int*  pTmp;

  Int   idx;


  pTmpBuff = (Int*)xMalloc( Int, iWidth * iHeight );
  
  // pad the RL recon image
  for ( y = 0;  y < iHeightRl; ++y ) 
  { 
    *(pRecRl-2) = *(pRecRl-1) = *pRecRl; 
	*(pRecRl+iWidthRl) = *(pRecRl+iWidthRl+1) = *(pRecRl+iWidthRl-1); 
	pRecRl += iStrideRl;
  }
  pRecRl = getPicYuvRec()->getLumaAddr();
  memcpy(pRecRl-iStrideRl-2, pRecRl-2, (iWidthRl+4)*sizeof(Pel));
  memcpy(pRecRl-2*iStrideRl-2, pRecRl-2, (iWidthRl+4)*sizeof(Pel));
  memcpy(pRecRl+(iHeightRl)*iStrideRl-2, pRecRl+(iHeightRl-1)*iStrideRl-2, (iWidthRl+4)*sizeof(Pel));
  memcpy(pRecRl+(iHeightRl+1)*iStrideRl-2, pRecRl+(iHeightRl-1)*iStrideRl-2, (iWidthRl+4)*sizeof(Pel));

  // interpolate
  pRecRl = pRecRl - 2*iStrideRl - 2;
  for ( y = 0, pTmp = pTmpBuff; y < iHeight; y += 2 )
  {
    for ( x = 0; x < iWidthRl+4; x++ )
      pTmp[x] = -pRecRl[x] +8*pRecRl[iStrideRl+x] +28*pRecRl[2*iStrideRl+x] -3*pRecRl[3*iStrideRl+x];
	pRecRl += iStrideRl;
    pTmp += iWidth;
    for ( x = 0; x < iWidthRl+4; x++ )
      pTmp[x] = -3*pRecRl[x] +28*pRecRl[iStrideRl+x] +8*pRecRl[2*iStrideRl+x] -pRecRl[3*iStrideRl+x];
    pTmp += iWidth;
  }

  for ( y = 0, pTmp = pTmpBuff; y < iHeight; y++, pTmp += iWidth, pRec += iStride )
  {
    for ( x = 0, idx = 0; x < iWidth; )
	{
      tmp = (-pTmp[idx] + 8*pTmp[idx+1] +28*pTmp[idx+2] -3*pTmp[idx+3] +(1<<9)) >> 10; 
	  pRec[x] = Clip(tmp);
	  idx += 1; x += 1;
      tmp = (-3*pTmp[idx] + 28*pTmp[idx+1] +8*pTmp[idx+2] -pTmp[idx+3] +(1<<9)) >> 10;
	  pRec[x] = Clip(tmp);
	  x += 1;
	}
  }



  pRecRl  = getPicYuvRec()->getCbAddr();
  pRec    = pcPic->getScaledRlPicRec()->getCbAddr();
  iWidthRl  >>= 1;
  iHeightRl >>= 1;
  iStrideRl >>= 1;
  iWidth    >>= 1;
  iHeight   >>= 1;
  iStride   >>= 1;

  for ( y = 0;  y < iHeightRl; ++y ) 
  { 
    *(pRecRl-1) = *pRecRl; 
	*(pRecRl+iWidthRl) = *(pRecRl+iWidthRl-1); 
	pRecRl += iStrideRl;
  }
  pRecRl = getPicYuvRec()->getCbAddr();
  memcpy(pRecRl-iStrideRl-1, pRecRl-1, (iWidthRl+2)*sizeof(Pel));
  memcpy(pRecRl+(iHeightRl)*iStrideRl-1, pRecRl+(iHeightRl-1)*iStrideRl-1, (iWidthRl+2)*sizeof(Pel));

  pRecRl = pRecRl - iStrideRl - 1;
  for ( y = 0, pTmp = pTmpBuff; y < iHeight; y += 2 )
  {
    for ( x = 0; x < iWidthRl+2; x++ )
      pTmp[x] = 4*pRecRl[x] +12*pRecRl[iStrideRl+x];
	pRecRl += iStrideRl;
    pTmp += iWidth;
    for( x = 0; x < iWidthRl+2; x++ )
      pTmp[x] = 12*pRecRl[x] +4*pRecRl[iStrideRl+x];
    pTmp += iWidth;
  }

  for ( y = 0, pTmp = pTmpBuff; y < iHeight; y++, pTmp += iWidth, pRec += iStride )
  {
    for ( x = 0, idx = 0; x < iWidth; )
	{
      tmp = (4*pTmp[idx] + 12*pTmp[idx+1] +(1<<7)) >> 8;  
	  pRec[x] = Clip(tmp);
	  idx += 1; x += 1;
      tmp = (12*pTmp[idx] + 4*pTmp[idx+1] +(1<<7)) >> 8;
	  pRec[x] = Clip(tmp);
	  x += 1;
	}
  }


  pRecRl  = getPicYuvRec()->getCrAddr();
  pRec    = pcPic->getScaledRlPicRec()->getCrAddr();

  // pad the RL recon image
  for ( y = 0;  y < iHeightRl; ++y ) 
  { 
	*(pRecRl-1) = *pRecRl; 
	*(pRecRl+iWidthRl) = *(pRecRl+iWidthRl-1); 
	pRecRl += iStrideRl;
  }
  pRecRl = getPicYuvRec()->getCrAddr();
  memcpy(pRecRl-iStrideRl-1, pRecRl-1, (iWidthRl+2)*sizeof(Pel));
  memcpy(pRecRl+(iHeightRl)*iStrideRl-1, pRecRl+(iHeightRl-1)*iStrideRl-1, (iWidthRl+2)*sizeof(Pel));

  pRecRl = pRecRl - iStrideRl - 1;
  for ( y = 0, pTmp = pTmpBuff; y < iHeight; y += 2 )
  {
    for ( x = 0; x < iWidthRl+2; x++ )
      pTmp[x] = 4*pRecRl[x] +12*pRecRl[iStrideRl+x];
	pRecRl += iStrideRl;
    pTmp += iWidth;
    for ( x = 0; x < iWidthRl+2; x++ )
      pTmp[x] = 12*pRecRl[x] +4*pRecRl[iStrideRl+x];
    pTmp += iWidth;
  }

  for ( y = 0, pTmp = pTmpBuff; y < iHeight; y++, pTmp += iWidth, pRec += iStride )
  {
    for ( x = 0, idx = 0; x < iWidth; )
	{
      tmp = (4*pTmp[idx] + 12*pTmp[idx+1] +(1<<7)) >> 8;
	  pRec[x] = Clip(tmp);
	  idx += 1; x += 1;
      tmp = (12*pTmp[idx] + 4*pTmp[idx+1] +(1<<7)) >> 8;
	  pRec[x] = Clip(tmp);
	  x += 1;
	}
  }


  xFree( pTmpBuff ); 
}

#endif
